
/**Copyright (c) 2009, Ioannis (John) Chtzikonstantinou, All rights reserved.
 ***http://prototy.blogspot.com
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 	- Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 	- Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 */
void updatelinePoints() {

  linePoints.clear();

  float mx, my, Mx, My;

  if (lineSegments.size()>0) {
    // define search borders
    lineSegment ls=(lineSegment)lineSegments.get(0);
    mx=min(ls.p1.x-ls.W,ls.p2.x-ls.W);
    my=min(ls.p1.y-ls.W,ls.p2.y-ls.W);
    Mx=max(ls.p1.x+ls.W,ls.p2.x+ls.W);
    My=max(ls.p1.y+ls.W,ls.p2.y+ls.W);
    for (int i=1, j=lineSegments.size();i<j;i++) {
      ls=(lineSegment)lineSegments.get(i);
      mx=min(mx,min(ls.p1.x-ls.W,ls.p2.x-ls.W));
      my=min(my,min(ls.p1.y-ls.W,ls.p2.y-ls.W));
      Mx=max(Mx,max(ls.p1.x+ls.W,ls.p2.x+ls.W));
      My=max(My,max(ls.p1.y+ls.W,ls.p2.y+ls.W));
    }
    // define active points by checking relative distances to line segments
    for (float l=mx;l<Mx;l++) {
      for (float k=my;k<My;k++) {
        linePoint2 ap=null;
        float d=-1;
        for (int i=0,j=lineSegments.size();i<j;i++) {
          ls=(lineSegment)lineSegments.get(i);
          float td=abs(dTS(ls.p1,ls.p2,new Point2(l,k)));
          if (td<ls.W && ( td<d || d==-1)) {
            int index=fluidSolver.getIndexForCellPosition((int)l,(int)k);
            Point2 P=new Point2(l,k);  // Point
            Point2 Pp=pTS(ls.p1,ls.p2,P);  //Projection
            Point2 Pn=P.minus(Pp);  // Normal
            ap=new linePoint2(index,Pn);
            d=td;
          }
        }
        if (ap!=null) {
          linePoints.add(ap);
        }
      }
    }
  }
}
void updateThrustPoints() {

  thrustPoints.clear();

  float mx, my, Mx, My;

  if (thrustSegments.size()>0) {
    // define search borders
    thrustSegment ts=(thrustSegment)thrustSegments.get(0);
    mx=min(ts.p1.x-ts.W,ts.p2.x-ts.W);
    my=min(ts.p1.y-ts.W,ts.p2.y-ts.W);
    Mx=max(ts.p1.x+ts.W,ts.p2.x+ts.W);
    My=max(ts.p1.y+ts.W,ts.p2.y+ts.W);
    for (int i=1, j=thrustSegments.size();i<j;i++) {
      ts=(thrustSegment)thrustSegments.get(i);
      mx=min(mx,min(ts.p1.x-ts.W,ts.p2.x-ts.W));
      my=min(my,min(ts.p1.y-ts.W,ts.p2.y-ts.W));
      Mx=max(Mx,max(ts.p1.x+ts.W,ts.p2.x+ts.W));
      My=max(My,max(ts.p1.y+ts.W,ts.p2.y+ts.W));
    }
    // define active points by checking relative distances to line segments
    for (float l=mx;l<Mx;l++) {
      for (float k=my;k<My;k++) {
        thrustPoint2 ap=null;
        float d=-1;
        for (int i=0,j=thrustSegments.size();i<j;i++) {
          ts=(thrustSegment)thrustSegments.get(i);
          float td=abs(dTS(ts.p1,ts.p2,new Point2(l,k)));
          if (td<ts.W && ( td<d || d==-1)) {
            int index=fluidSolver.getIndexForCellPosition((int)l,(int)k);
            Point2 P=new Point2(l,k);  // Point
            ap=new thrustPoint2(index,ts.Pth);
            d=td;
          }
        }
        if (ap!=null) {
          thrustPoints.add(ap);
        }
      }
    }
  }
}
// Update solver and draw
void updateSolver() {

  for (int i=0, j=linePoints.size();i<j;i++) {
    linePoint2 p=(linePoint2)linePoints.get(i);
    p.update();
  }
  for (int i=0, j=thrustPoints.size();i<j;i++) {
    thrustPoint2 p=(thrustPoint2)thrustPoints.get(i);
    p.update();
  }
  fluidSolver.update();
  if (tunnelFlow) addTunnelFlow();
}
//Default Wind Tunnel Flow
void addTunnelFlow() {
  for (int i=0;i<FLUID_WIDTH*height/width;i++) {
    int index = fluidSolver.getIndexForCellPosition((int)FLUID_WIDTH-2,i);
    fluidSolver.u[index] -= 59.0;
  }
  for (int i=0;i<FLUID_WIDTH*height/width;i++) {
    int index = fluidSolver.getIndexForCellPosition(0,i);
    fluidSolver.u[index] -= 28.0;
  }
  for (int i=0;i<30;i++) {
    int index = fluidSolver.getIndexForCellPosition((int)FLUID_WIDTH-2,i);
  }
}




