import processing.core.*; 
import processing.xml.*; 

import msafluid.*; 
import processing.opengl.*; 
import javax.media.opengl.*; 
import java.util.Vector; 
import processing.dxf.*; 
import peasy.*; 

import java.applet.*; 
import java.awt.*; 
import java.awt.image.*; 
import java.awt.event.*; 
import java.io.*; 
import java.net.*; 
import java.text.*; 
import java.util.*; 
import java.util.zip.*; 
import java.util.regex.*; 

public class TN7 extends PApplet {


/**Copyright (c) 2009, Ioannis (John) Chtzikonstantinou, ?All rights reserved.
 ***http://prototy.blogspot.com
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 	- Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 	- Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 */






static final float FLUID_WIDTH = 180;
static int noCells;
static float invWidth, invHeight;    // inverse of screen dimensions

MSAFluidSolver2D fluidSolver;

Vector lineSegments = new Vector();
Vector linePoints=new Vector();
Vector thrustSegments = new Vector();
Vector thrustPoints=new Vector();

float[] Velocities;

// Real Values
int tx=0;
int ty=0;
//Simulation Values
float mX=0;
float mY=0;
float tX=0;
float tY=0;

int drawMode=0;
boolean drawObj=true;
boolean tunnelFlow=false;
boolean visMode=false;

static PImage imgVis;
static PImage imgFlow;
static PImage img2d;

public void setup() {
  size(1024, 450, OPENGL);    // use OPENGL rendering for bilinear filtering on texture
  frameRate(24);
  hint( ENABLE_OPENGL_4X_SMOOTH );    // Turn on 4X antialiasing
  this.frame.setTitle("TN6");
  background(0);
  // create fluid and set options
  invWidth=1.0f/width;
  invHeight=1.0f/height;
  fluidSolver = new MSAFluidSolver2D((int)(FLUID_WIDTH), (int)(FLUID_WIDTH * height/width));
  fluidSolver.setSolverIterations(10).setDeltaT(.00006f).setVisc(0.35f);

  imgVis = loadImage("vis.jpg");
  imgFlow = loadImage("flow.jpg");
  img2d = loadImage("2d.jpg");
  
  Velocities = new float[fluidSolver.u.length];

  setupCam();
  noCursor();
}

public void draw() {
  mX=(mouseX*FLUID_WIDTH/width);
  mY=(mouseY*FLUID_WIDTH/width);

  updateSolver();
  if (!visMode) {
    rDraw();
  } 
  else {
    draw3DField();
  }
}



/*
 * DistancePointSegmentExample, calculate distance to line
 * Copyright (C) 2008 Pieter Iserbyt <pieter.iserbyt@gmail.com>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Example implementation of "Minimum Distance between a Point and a Line" as
 * described by Paul Bourke on
 * See http://local.wasp.uwa.edu.au/~pbourke/geometry/pointline/.
 */

//Modified to return projection to segment and distance to segment
//by Ioannis (John) Chatzikonstantinou in May 2009
// Also simplified
// http://prototy.blogspot.com

/**
 * Returns the projection of p3 to the segment defined by p1,p2;
 * 
 * @param p1
 *                First point of the segment
 * @param p2
 *                Second point of the segment
 * @param p3
 *                Point of which we want to know the projection on the segment
 *                defined by p1,p2
 * @return The projection of p3 to the segment defined by p1,p2
 */
public static Point2 pTS(Point2 p1, Point2 p2, Point2 p3) {

  final float xDelta = p2.x-p1.x;
  final float yDelta = p2.y-p1.y;
  final Point2 closestPoint;

  if ((xDelta == 0) && (yDelta == 0)) {
    closestPoint=p1;
  } 
  else {

    final float u = ((p3.x - p1.x)*xDelta + (p3.y - p1.y)*yDelta) / (xDelta*xDelta + yDelta*yDelta);

    if (u < 0) {
      closestPoint = p1;
    } 
    else if (u > 1) {
      closestPoint = p2;
    }
    else {
      closestPoint = new Point2(p1.x + u*xDelta, p1.y + u*yDelta);
    }
  }
  return(closestPoint);
}

/**
 * Returns the projection of p3 to the line defined by p1,p2;
 * 
 * @param p1
 *                First point of the segment
 * @param p2
 *                Second point of the segment
 * @param p3
 *                Point to which we want to know the distance of the line
 *                defined by p1,p2
 * @return The projection of p3 to the line defined by p1,p2
 */
public static Point2 pTL(Point2 p1, Point2 p2, Point2 p3) {

  final float xDelta = p2.x - p1.x;
  final float yDelta = p2.y - p1.y;
  final Point2 closestPoint;

  if ((xDelta == 0) && (yDelta == 0)) {
    closestPoint=new Point2(0,0);
  } 
  else {

    final float u = ((p3.x - p1.x) * xDelta + (p3.y - p1.y) * yDelta) / (xDelta * xDelta + yDelta * yDelta);

    closestPoint = new Point2(p1.x + u * xDelta, p1.y + u * yDelta);
  }
  return(closestPoint);
}
/**
 * Returns the distance of p3 to the segment defined by p1,p2;
 * 
 * @param p1
 *                First point of the segment
 * @param p2
 *                Second point of the segment
 * @param p3
 *                Point to which we want to know the distance of the segment
 *                defined by p1,p2
 * @return The distance of p3 to the segment defined by p1,p2
 */
public static float dTS(Point2 p1, Point2 p2, Point2 p3) {

  final Point2 closestPoint=pTS(p1, p2, p3);
  return closestPoint.distance(p3);
}
/**
 * Returns the distance of p3 to the line defined by p1,p2;
 * 
 * @param p1
 *                First point of the segment
 * @param p2
 *                Second point of the segment
 * @param p3
 *                Point to which we want to know the distance of the segment
 *                defined by p1,p2
 * @return The distance of p3 to the segment defined by p1,p2
 */
public static float dTL(Point2 p1, Point2 p2, Point2 p3) {

  final Point2 closestPoint=pTL(p1, p2, p3);
  return closestPoint.distance(p3);
}



/**Copyright (c) 2009, Ioannis (John) Chtzikonstantinou, ?All rights reserved.
 ***http://prototy.blogspot.com
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 	- Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 	- Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 */


public void export(String file) {
   beginRaw(DXF, file); // Start recording to the file
   final float f=width/FLUID_WIDTH;
  final float ff=0.20f*f;
  final float rt=5;
 
  for (int i=0;i<Velocities.length;i++) {
    Velocities[i]=dist(0,0,fluidSolver.u[i],fluidSolver.v[i]);
  }
  for (int j=0,l=(int)(FLUID_WIDTH*height/width)-1;j<l;j++) {
    for (int i=5,k=(int)FLUID_WIDTH-(int)(125/f)-1;i<k;i++) {
      int index0=fluidSolver.getIndexForCellPosition(i,j);
      int index1=fluidSolver.getIndexForCellPosition(i+1,j);
      int index2=fluidSolver.getIndexForCellPosition(i,j+1);
      int index3=fluidSolver.getIndexForCellPosition(i+1,j+1);
      final float f0=rt*Velocities[index0];
      final float f1=rt*Velocities[index1];
      final float f2=rt*Velocities[index2];
      final float f3=rt*Velocities[index3];
      beginShape(TRIANGLE_FAN);
      vertex(f*i,f*j,f0);
      vertex(f*(i+1),f*j,f1);
      vertex(f*(i+1),f*(j+1),f3);
      vertex(f*i,f*(j+1),f2);
      endShape();
      beginShape(TRIANGLE_FAN);
      vertex(f*i,f*j,-f0);
      vertex(f*(i+1),f*j,-f1);
      vertex(f*(i+1),f*(j+1),-f3);
      vertex(f*i,f*(j+1),-f2);
      endShape();
    }
  }
    endRaw();
}


/**Copyright (c) 2009, Ioannis (John) Chtzikonstantinou, All rights reserved.
 ***http://prototy.blogspot.com
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 	- Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 	- Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 */
Point2 p1,p2;

public void mousePressed() {
  if (!visMode) {
  if (drawMode==0) {
    tx = mouseX;
    ty = mouseY;
    tX=(tx*FLUID_WIDTH/width);
    tY=(ty*FLUID_WIDTH/width);
  }
  }
}
public void mouseReleased() {
  if (!visMode) {
  if (abs(tx - mouseX)<3 && abs(ty - mouseY)<3 && drawMode==0) {
    drawMode=1;
  } 
  else if (drawMode==1) {
    p1=new Point2(tX,tY);
    p2=new Point2(mX,mY);
    lineSegments.add(new lineSegment(p1,p2));
    updatelinePoints();
    drawMode=0;
  } 
  else if (drawMode==2) {
    Point2 p3=new Point2(mX,mY);
    thrustSegments.add(new thrustSegment(p1,p2,p3));
    updateThrustPoints();
    drawMode=0;
  }
  else {
    p1=new Point2(tX,tY);
    p2=new Point2(mX,mY);
    drawMode=2;
  }
  }
}
public void keyPressed() {
  if (keyCode==' ') {
    tunnelFlow=!tunnelFlow;
  }
  else if (keyCode=='3') {
    drawObj=!drawObj;
  }
  else if (keyCode=='R') {
    fluidSolver.reset();
    lineSegments=new Vector();
    linePoints=new Vector();
    thrustSegments=new Vector();
    thrustPoints=new Vector();
  } else if (keyCode=='1' && visMode) {
    visMode=false;
    state3D=cam.getState();
    cam.setState(state2D,0);
    cam.setMouseControlled(false);
    noCursor();
  } else if (keyCode=='2' && !visMode) {
    visMode=true;
    state2D=cam.getState();
    cam.setState(state3D,0);
    cam.setMouseControlled(true);
    cursor();
  }
  else if (keyCode=='X') {
    export("TN7_out.dxf");
  }
}



/**Copyright (c) 2009, Ioannis (John) Chtzikonstantinou, All rights reserved.
 ***http://prototy.blogspot.com
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 	- Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 	- Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 */

class lineSegment {

  Point2 p1,p2;

  float W=1.81f;

  lineSegment(Point2 p1in, Point2 p2in) {
    p1=p1in;
    p2=p2in;
  }
  public void draw() {
    float r=width/FLUID_WIDTH;
    float w=W*2;
    stroke(255);
    strokeWeight(w);
    line(p1.x*r,p1.y*r,p2.x*r,p2.y*r);
    noStroke();
    fill(255);
    ellipse(p1.x*r,p1.y*r,w,w);
    ellipse(p2.x*r,p2.y*r,w,w);
  }
}

public final class linePoint2 {
  int index;
  Point2 Pn;
  linePoint2(int indexin,Point2 pnormalin) {
    index=indexin;
    Pn=pnormalin;
  }
  public void update() {
    Point2 U=new Point2(fluidSolver.u[index],fluidSolver.v[index]);
    Point2 Un=pTL(new Point2(0,0),Pn,U);// project U to normal
    Un.mult(-1.00f);
    fluidSolver.u[index]+=Un.x;
    fluidSolver.v[index]+=Un.y;
  }
}

public static final class Point2 {
  float x,y;
  Point2(float xin, float yin) {
    x=xin;
    y=yin;
  }
  public float distance(Point2 pin) {
    float dx=pin.x-x;
    float dy=pin.y-y;
    return sqrt(dx*dx+dy*dy);
  }
  public Point2 plus(Point2 pin) {
    return new Point2(x+pin.x,y+pin.y);
  }
  public Point2 minus(Point2 pin) {
    return new Point2(x-pin.x,y-pin.y);
  }
  public Point2 mult(float i) {
    return new Point2(x*=i, y*=i);
  }
  public Point2 cp() {
    return new Point2(x,y);
  }
  public Point2 lperp() {
    return new Point2(-y,x);
  }
  public Point2 rperp() {
    return new Point2(y,-x);
  }
  public Point2 n() {
    final float m=sqrt(x*x+y*y);
    return new Point2(x/m,y/m);
  }
  public void inherit(Point2 pin) {
    x=pin.x;
    y=pin.y;
  }
  public float dot(Point2 pin) {
    return x*pin.x+y*pin.y;
  }
}



/**Copyright (c) 2009, Ioannis (John) Chtzikonstantinou, All rights reserved.
 ***http://prototy.blogspot.com
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 	- Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 	- Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 */

class thrustSegment {

  Point2 p1,p2,p3;

  Point2 Pth;

  final float W=1.81f;
  final float r=width/FLUID_WIDTH;
  final float w=W*2;
  final float staticMult=.1f;

  thrustSegment(Point2 p1in, Point2 p2in, Point2 p3in) {
    p1=p1in;
    p2=p2in;
    p3=p3in;
    Pth=p3.minus(pTL(p1,p2,p3)).mult(staticMult);
  }
  public void draw() {
    stroke(130);
    strokeWeight(10);
    line(p1.x*r,p1.y*r,p2.x*r,p2.y*r);
  }
}

public final class thrustPoint2 {

  int index;
  Point2 Pth;

  thrustPoint2(int indexin,Point2 Pthin) {
    index=indexin;
    Pth=Pthin;
  }
  public void update() {
    fluidSolver.u[index]+=Pth.x;
    fluidSolver.v[index]+=Pth.y;
  }
}



/**Copyright (c) 2009, Ioannis (John) Chtzikonstantinou, All rights reserved.
 ***http://prototy.blogspot.com
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 	- Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 	- Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 */
public void updatelinePoints() {

  linePoints.clear();

  float mx, my, Mx, My;

  if (lineSegments.size()>0) {
    // define search borders
    lineSegment ls=(lineSegment)lineSegments.get(0);
    mx=min(ls.p1.x-ls.W,ls.p2.x-ls.W);
    my=min(ls.p1.y-ls.W,ls.p2.y-ls.W);
    Mx=max(ls.p1.x+ls.W,ls.p2.x+ls.W);
    My=max(ls.p1.y+ls.W,ls.p2.y+ls.W);
    for (int i=1, j=lineSegments.size();i<j;i++) {
      ls=(lineSegment)lineSegments.get(i);
      mx=min(mx,min(ls.p1.x-ls.W,ls.p2.x-ls.W));
      my=min(my,min(ls.p1.y-ls.W,ls.p2.y-ls.W));
      Mx=max(Mx,max(ls.p1.x+ls.W,ls.p2.x+ls.W));
      My=max(My,max(ls.p1.y+ls.W,ls.p2.y+ls.W));
    }
    // define active points by checking relative distances to line segments
    for (float l=mx;l<Mx;l++) {
      for (float k=my;k<My;k++) {
        linePoint2 ap=null;
        float d=-1;
        for (int i=0,j=lineSegments.size();i<j;i++) {
          ls=(lineSegment)lineSegments.get(i);
          float td=abs(dTS(ls.p1,ls.p2,new Point2(l,k)));
          if (td<ls.W && ( td<d || d==-1)) {
            int index=fluidSolver.getIndexForCellPosition((int)l,(int)k);
            Point2 P=new Point2(l,k);  // Point
            Point2 Pp=pTS(ls.p1,ls.p2,P);  //Projection
            Point2 Pn=P.minus(Pp);  // Normal
            ap=new linePoint2(index,Pn);
            d=td;
          }
        }
        if (ap!=null) {
          linePoints.add(ap);
        }
      }
    }
  }
}
public void updateThrustPoints() {

  thrustPoints.clear();

  float mx, my, Mx, My;

  if (thrustSegments.size()>0) {
    // define search borders
    thrustSegment ts=(thrustSegment)thrustSegments.get(0);
    mx=min(ts.p1.x-ts.W,ts.p2.x-ts.W);
    my=min(ts.p1.y-ts.W,ts.p2.y-ts.W);
    Mx=max(ts.p1.x+ts.W,ts.p2.x+ts.W);
    My=max(ts.p1.y+ts.W,ts.p2.y+ts.W);
    for (int i=1, j=thrustSegments.size();i<j;i++) {
      ts=(thrustSegment)thrustSegments.get(i);
      mx=min(mx,min(ts.p1.x-ts.W,ts.p2.x-ts.W));
      my=min(my,min(ts.p1.y-ts.W,ts.p2.y-ts.W));
      Mx=max(Mx,max(ts.p1.x+ts.W,ts.p2.x+ts.W));
      My=max(My,max(ts.p1.y+ts.W,ts.p2.y+ts.W));
    }
    // define active points by checking relative distances to line segments
    for (float l=mx;l<Mx;l++) {
      for (float k=my;k<My;k++) {
        thrustPoint2 ap=null;
        float d=-1;
        for (int i=0,j=thrustSegments.size();i<j;i++) {
          ts=(thrustSegment)thrustSegments.get(i);
          float td=abs(dTS(ts.p1,ts.p2,new Point2(l,k)));
          if (td<ts.W && ( td<d || d==-1)) {
            int index=fluidSolver.getIndexForCellPosition((int)l,(int)k);
            Point2 P=new Point2(l,k);  // Point
            ap=new thrustPoint2(index,ts.Pth);
            d=td;
          }
        }
        if (ap!=null) {
          thrustPoints.add(ap);
        }
      }
    }
  }
}
// Update solver and draw
public void updateSolver() {

  for (int i=0, j=linePoints.size();i<j;i++) {
    linePoint2 p=(linePoint2)linePoints.get(i);
    p.update();
  }
  for (int i=0, j=thrustPoints.size();i<j;i++) {
    thrustPoint2 p=(thrustPoint2)thrustPoints.get(i);
    p.update();
  }
  fluidSolver.update();
  if (tunnelFlow) addTunnelFlow();
}
//Default Wind Tunnel Flow
public void addTunnelFlow() {
  for (int i=0;i<FLUID_WIDTH*height/width;i++) {
    int index = fluidSolver.getIndexForCellPosition((int)FLUID_WIDTH-2,i);
    fluidSolver.u[index] -= 59.0f;
  }
  for (int i=0;i<FLUID_WIDTH*height/width;i++) {
    int index = fluidSolver.getIndexForCellPosition(0,i);
    fluidSolver.u[index] -= 28.0f;
  }
  for (int i=0;i<30;i++) {
    int index = fluidSolver.getIndexForCellPosition((int)FLUID_WIDTH-2,i);
  }
}





/**Copyright (c) 2009, Ioannis (John) Chtzikonstantinou, All rights reserved.
 ***http://prototy.blogspot.com
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 	- Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 	- Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 */

// Visualize Velocity Field

public void drawVelocityField() {
  final float f=width/FLUID_WIDTH;
  final float ff=0.80f*f;
  noFill();
  strokeWeight(1);
  for (int j=0,l=(int)(FLUID_WIDTH*height/width);j<l;j++) {
    for (int i=0,k=(int)FLUID_WIDTH-(int)(125/f);i<k;i++) {
      int index=fluidSolver.getIndexForCellPosition(i,j);
      float v2=fluidSolver.u[index]*fluidSolver.u[index]+fluidSolver.v[index]*fluidSolver.v[index];
      final float f2=2.9f+0.03f*v2;
      stroke(255,22);
      line(i*f+random(f2)-f2*0.50f,j*f+random(f2)-f2*0.50f,i*f+fluidSolver.u[index]*ff+random(f2)-f2*0.50f,
      j*f+fluidSolver.v[index]*ff+random(f2)-f2*0.50f);
    }
  }
}

public void rDraw() {
background(0);
  drawVelocityField();
  if (drawObj) {
    for (int i=0,j=lineSegments.size();i<j;i++) {
      lineSegment l=(lineSegment)lineSegments.get(i);
      l.draw();
    }
    for (int i=0,j=thrustSegments.size();i<j;i++) {
      thrustSegment t=(thrustSegment)thrustSegments.get(i);
      t.draw();
    }
  }
  noFill();
  stroke(220);
  strokeWeight(1);
  line(mouseX,0,mouseX,width-50);
  line(0,mouseY,width,mouseY);
  if (drawMode==1) {
    noFill();
    stroke(225);
    strokeWeight(3);
    line(tx,ty,mouseX,mouseY);
  }
  else if (drawMode==2) {
    stroke(100);
    strokeWeight(10);
    float r=width/FLUID_WIDTH;
    line(p1.x*r,p1.y*r,p2.x*r,p2.y*r);
  }
  else if (drawMode==0 && mousePressed) {
    noFill();
    stroke(100);
    strokeWeight(3);
    line(tx,ty,mouseX,mouseY);
  }
  noStroke();
  fill(0);
  rect(width-120,0,120,height);
  if (tunnelFlow) {
    image(imgFlow, width-120, height-30); 
  }
  if (drawObj) {
    image(imgVis, width-100, height-30); 
  }
  if (!visMode) {
    image(img2d, width-80,height-30);
  }
}

/**Copyright (c) 2009, Ioannis (John) Chtzikonstantinou, ?All rights reserved.
 ***http://prototy.blogspot.com
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 	- Redistributions of source code must retain the above copyright 
 * notice, this list of conditions and the following disclaimer.
 * 	- Redistributions in binary form must reproduce the above copyright 
 * notice, this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT 
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 */



PeasyCam cam;
CameraState state2D;
CameraState state3D;

public void setupCam() {
  cam = new PeasyCam(this, 0,0,0,800);
  cam.setMouseControlled(false);
  cam.setMinimumDistance(50);
  cam.setMaximumDistance(1500);
  cam.pan(width/2,height/2);
  cam.setDistance(390);
  state2D=cam.getState();
  state3D=cam.getState();
}


public void draw3DField() {
  final float f=width/FLUID_WIDTH;
  final float ff=0.20f*f;
  final float rt=5;
  
  lights();
  
  background(0);
  noStroke();
  fill(200);
  for (int i=0;i<Velocities.length;i++) {
    Velocities[i]=dist(0,0,fluidSolver.u[i],fluidSolver.v[i]);
  }
  for (int j=0,l=(int)(FLUID_WIDTH*height/width)-1;j<l;j++) {
    for (int i=5,k=(int)FLUID_WIDTH-(int)(125/f)-1;i<k;i++) {
      int index0=fluidSolver.getIndexForCellPosition(i,j);
      int index1=fluidSolver.getIndexForCellPosition(i+1,j);
      int index2=fluidSolver.getIndexForCellPosition(i,j+1);
      int index3=fluidSolver.getIndexForCellPosition(i+1,j+1);
      final float f0=rt*Velocities[index0];
      final float f1=rt*Velocities[index1];
      final float f2=rt*Velocities[index2];
      final float f3=rt*Velocities[index3];
      beginShape(TRIANGLE_FAN);
      vertex(f*i,f*j,f0);
      vertex(f*(i+1),f*j,f1);
      vertex(f*(i+1),f*(j+1),f3);
      vertex(f*i,f*(j+1),f2);
      endShape();
      beginShape(TRIANGLE_FAN);
      vertex(f*i,f*j,-f0);
      vertex(f*(i+1),f*j,-f1);
      vertex(f*(i+1),f*(j+1),-f3);
      vertex(f*i,f*(j+1),-f2);
      endShape();
    }
  }
}

  static public void main(String args[]) {
    PApplet.main(new String[] { "--bgcolor=#FFFFFF", "TN7" });
  }
}
